module genum;
import std.typetuple : indexOf;

//--------------------------------------------------------------
// exists!(T).inside!(X) は、X(のメンバ)* に型 T が
// あるとき true。ないとき false。
// とりあえず配列とクラスだけ考えることにする
//--------------------------------------------------------------

private
template exists(T)
{
   template inside(X) { const inside = rec!(X).value; }

   template rec(X, Visited...) {
      // 再帰ですでに一度Xを調べかけていたら
      // ループしてるのでここでは即座にfalseを返す
      static if( indexOf!(X, Visited) != -1 )
         const value = false;
      else
         // X == T の場合
         static if( is(X == T) )
            const value = true;
         // X が配列の場合
         else
         static if( is(X E == E[]) )
            const value = rec!(E, Visited).value;
         // X がクラスの場合
         else
         static if( is(X == class) )
         {
            // ちなみに
            // ここで inside_any というサブ関数が必要になるため
            // 「templateの唯一のメンバは省略記法が使える」
            // ルールを生かせないので rec!(...).rec ではなく
            // rec!(...).value という形で値を返しています。

            template inside_any() {
               const inside_any = false;
            }
            template inside_any(Y,YS...) {
               const inside_any = rec!(Y, Visited,X).value
                               || inside_any!(YS);
            }
            const value = inside_any!(typeof(X.tupleof));
         }
         // その他（※本当は構造体や連想配列も考えるべき）
         else
            const value = false;
   }
}

//--------------------------------------------------------------
// x の中で T がある部分を再帰的にたどる
// T そのものを見つけたら fn を呼ぶ
//--------------------------------------------------------------

private
void each_impl(T, X, F)( X x, F fn )
{
    static if( exists!(T).inside!(X) )
       static if( is(X == T) )
          fn( x );
       else if( x !is null )
          static if( is(X E == E[]) )
             foreach(elem; x)
                each_impl!(T)( elem, fn );
          else
             foreach(field; x.tupleof)
                each_impl!(T)( field, fn );
}

// Shadow Type を使ったバージョン
private
void each_impl_shadow(ST, SX, X, F)( X x, F fn )
{
   static if( exists!(ST).inside!(SX) )
      static if( is(SX == ST) )
         fn( x );
      else if( x !is null )
         static if( is(SX SE == SE[]) )
            foreach(elem; x)
               each_impl_shadow!(ST, SE)( elem, fn );
         else
            foreach(i, field; x.tupleof)
               each_impl_shadow!
                 (ST, typeof(SX.tupleof[i]))( field, fn );
}

//--------------------------------------------------------------
// cont の各要素に fn を適用
//--------------------------------------------------------------

public
void each(T, alias Cont, F)( Cont!(T) cont, F fn )
{
   //each_impl!(T)( cont, fn );

   class ShadowType {
      T t; alias t this; // できる限り型Tの振りをする

      // 基底クラスObjectに元々定義されているメンバがあると
      // tに転送する前にそちらに取られてしまうので、
      // オーバーロードして無理やり奪い取り返す
      int  opCmp(S)(S s)    { return 0; }   // 型だけが問題なので
      bool opEquals(S)(S s) { return false; } // 実装は何でもいい
   }
   each_impl_shadow!(ShadowType, Cont!(ShadowType))( cont, fn );
}
