
//------------------------------------------------------------------
// とりあえず配列用のイテレータだけ。
// 普通に全てのイテレータをラップするのを書くのも特に手間ではないのですが、
// そもそもDの標準ライブラリに配列用イテレータしかないし…
//------------------------------------------------------------------

class DependentIterator(alias arr)
{
	private typeof(arr.ptr) impl;
	this( typeof(arr.ptr) i ) { this.impl = i; }

	DependentIterator opPostInc() { impl++; return new DependentIterator(impl-1); }
	DependentIterator opPostDec() { impl--; return new DependentIterator(impl-1); }
	DependentIterator opAdd(int i){ return new DependentIterator(impl+i); }
	DependentIterator opSub(int i){ return new DependentIterator(impl-i); }
	void  opAddAssign(int i){ impl += i; }
	void  opSubAssign(int i){ impl -= i; }
	auto  opStar()    { return *impl; }
	int   opSub(DependentIterator rhs){ return impl - rhs.impl; }
	bool  opEquals(DependentIterator rhs) { return impl == rhs.impl; }
	int   opCmp(DependentIterator rhs) { return impl < rhs.impl ? -1 : impl == rhs.impl ? 0 : +1; }
}

auto begin(alias arr)()
{
	return new DependentIterator!(arr)(arr.ptr);
}

auto end(alias arr)()
{
	return new DependentIterator!(arr)(arr.ptr + arr.length);
}


//------------------------------------------------------------------
// 例
//------------------------------------------------------------------

import std.stdio;
void main() { writeln("<100% Ok>"); }

void for_each(I,F)(I s, I e, F f)
{
	for(; s!=e; ++s)
		f(*s);
}

unittest
{
	int[] a = [1,2,3];
	int[] b = [4,5,6];

	assert(  is(typeof(for_each( begin!(a), end!(a), (int n){} ))) );
	assert(  is(typeof(for_each( begin!(b), end!(b), (int n){} ))) );
	assert( !is(typeof(for_each( begin!(a), end!(b), (int n){} ))) );

	int sa = 0;
	for_each( begin!(a), end!(a), (int n){sa+=n;} );
	assert( sa == 6 );

	int sb = 1;
	for_each( begin!(b), end!(b), (int n){sb*=n;} );
	assert( sb == 120 );
}
